/*
 * Jalview - A Sequence Alignment Editor and Viewer (Version 2.8.2)
 * Copyright (C) 2014 The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package MCview;

import java.io.*;
import java.util.*;

import java.awt.*;

import jalview.analysis.AlignSeq;
import jalview.datamodel.*;
import jalview.io.FileParse;
import jalview.util.MessageManager;

public class PDBfile extends jalview.io.AlignFile
{
  public Vector chains;

  public String id;

  /**
   * set to true to add chain alignment annotation as visible annotation.
   */
  boolean VisibleChainAnnotation = false;

  public PDBfile(String inFile, String inType) throws IOException
  {
    super(inFile, inType);
  }

  public PDBfile(FileParse source) throws IOException
  {
    super(source);
  }

  public String print()
  {
    return null;
  }

  public void parse() throws IOException
  {
    // TODO set the filename sensibly - try using data source name.
    id = safeName(getDataName());

    chains = new Vector();
    ArrayList<SequenceI> rna = new ArrayList<SequenceI>(), prot = new ArrayList<SequenceI>();
    PDBChain tmpchain;
    String line = null;
    boolean modelFlag = false;
    boolean terFlag = false;
    String lastID = "";

    int index = 0;
    String atomnam = null;
    try
    {
      while ((line = nextLine()) != null)
      {
        if (line.indexOf("HEADER") == 0)
        {
          if (line.length() > 62)
          {
            String tid;
            if (line.length() > 67)
            {
              tid = line.substring(62, 67).trim();
            }
            else
            {
              tid = line.substring(62).trim();
            }
            if (tid.length() > 0)
            {
              id = tid;
            }
            continue;
          }
        }
        // Were we to do anything with SEQRES - we start it here
        if (line.indexOf("SEQRES") == 0)
        {
        }

        if (line.indexOf("MODEL") == 0)
        {
          modelFlag = true;
        }

        if (line.indexOf("TER") == 0)
        {
          terFlag = true;
        }

        if (modelFlag && line.indexOf("ENDMDL") == 0)
        {
          break;
        }
        if (line.indexOf("ATOM") == 0
                || (line.indexOf("HETATM") == 0 && !terFlag))
        {
          terFlag = false;

          // Jalview is only interested in CA bonds????
          atomnam = line.substring(12, 15).trim();
          if (!atomnam.equals("CA") && !atomnam.equals("P"))
          {
            continue;
          }

          Atom tmpatom = new Atom(line);
          tmpchain = findChain(tmpatom.chain);
          if (tmpchain != null)
          {
            if (tmpatom.resNumIns.trim().equals(lastID))
            {
              // phosphorylated protein - seen both CA and P..
              continue;
            }
            tmpchain.atoms.addElement(tmpatom);
          }
          else
          {
            tmpchain = new PDBChain(id, tmpatom.chain);
            chains.addElement(tmpchain);
            tmpchain.atoms.addElement(tmpatom);
          }
          lastID = tmpatom.resNumIns.trim();
        }
        index++;
      }

      makeResidueList();
      makeCaBondList();

      if (id == null)
      {
        id = inFile.getName();
      }
      for (int i = 0; i < chains.size(); i++)
      {
        SequenceI dataset = ((PDBChain) chains.elementAt(i)).sequence;
        dataset.setName(id + "|" + dataset.getName());
        PDBEntry entry = new PDBEntry();
        entry.setId(id);
        entry.setProperty(new Hashtable());
        if (((PDBChain) chains.elementAt(i)).id != null)
        {
          entry.getProperty().put("CHAIN",
                  ((PDBChain) chains.elementAt(i)).id);
        }
        if (inFile != null)
        {
          entry.setFile(inFile.getAbsolutePath());
        }
        else
        {
          // TODO: decide if we should dump the datasource to disk
          entry.setFile(getDataName());
        }
        dataset.addPDBId(entry);
        SequenceI chainseq = dataset.deriveSequence(); // PDBChain objects
        // maintain reference to
        // dataset
        seqs.addElement(chainseq);
        if (isRNA(chainseq) == true)
        {
          rna.add(chainseq);
        }
        else
        {
          prot.add(chainseq);
        }

        AlignmentAnnotation[] chainannot = chainseq.getAnnotation();

        if (chainannot != null)
        {
          for (int ai = 0; ai < chainannot.length; ai++)
          {

            chainannot[ai].visible = VisibleChainAnnotation;
            annotations.addElement(chainannot[ai]);
          }
        }
      }
      if (rna.size() > 0)
        try
        {
          processPdbFileWithAnnotate3d(rna);
        } catch (Exception x)
        {
          System.err
                  .println("Exceptions when dealing with RNA in pdb file");
          x.printStackTrace();

        }
      ;
      if (prot.size() > 0)
        try
        {
          processPdbFileWithJmol(prot);
        } catch (Exception x)
        {
          System.err
                  .println("Exceptions when dealing with RNA in pdb file");
          x.printStackTrace();

        }
      ;
      if (prot.size() > 0)
        try
        {
          processPdbFileWithJmol(prot);
        } catch (Exception x)
        {
          System.err
                  .println("Exceptions when dealing with RNA in pdb file");
          x.printStackTrace();

        }
      ;
    } catch (OutOfMemoryError er)
    {
      System.out.println("OUT OF MEMORY LOADING PDB FILE");
      throw new IOException(MessageManager.getString("exception.outofmemory_loading_pdb_file"));
    } catch (NumberFormatException ex)
    {
      if (line != null)
      {
        System.err.println("Couldn't read number from line:");
        System.err.println(line);
      }
    }
  }

  private void processPdbFileWithJmol(ArrayList<SequenceI> prot)
          throws Exception
  {
    try
    {
      Class cl = Class.forName("jalview.ext.jmol.PDBFileWithJmol");
      if (cl != null)
      {
        Object jmf = cl.getConstructor(new Class[]
        { FileParse.class }).newInstance(new Object[]
        { new FileParse(getDataName(), type) });
        Alignment al = new Alignment((SequenceI[]) cl.getMethod(
                "getSeqsAsArray", new Class[]
                {}).invoke(jmf));
        cl.getMethod("addAnnotations", new Class[]
        { Alignment.class }).invoke(jmf, al);
        replaceMatchingSeqsWith(prot, al, AlignSeq.PEP);
      }
    } catch (ClassNotFoundException q)
    {
    }
  }

  private void processPdbFileWithAnnotate3d(ArrayList<SequenceI> rna)
          throws Exception
  {
    // System.out.println("this is a PDB format and RNA sequence");
    // note: we use reflection here so that the applet can compile and run
    // without the HTTPClient bits and pieces needed for accessing Annotate3D
    // web service
    try
    {
      Class cl = Class.forName("jalview.ws.jws1.Annotate3D");
      if (cl != null)
      {
        // TODO: use the PDB ID of the structure if one is available, to save
        // bandwidth and avoid uploading the whole structure to the service
        Object annotate3d = cl.getConstructor(new Class[]
        {}).newInstance(new Object[]
        {});
        AlignmentI al = ((AlignmentI) cl.getMethod("getRNAMLFor",
                new Class[]
                { FileParse.class }).invoke(annotate3d, new Object[]
        { new FileParse(getDataName(), type) }));
        replaceMatchingSeqsWith(rna, al, AlignSeq.DNA);
      }
    } catch (ClassNotFoundException x)
    {
      // ignore classnotfounds - occurs in applet
    }
    ;
  }

  private void replaceMatchingSeqsWith(ArrayList<SequenceI> ochains,
          AlignmentI al, String dnaOrProtein)
  {
    if (al != null && al.getHeight() > 0)
    {
      ArrayList<SequenceI> matches = new ArrayList<SequenceI>();
      ArrayList<AlignSeq> aligns = new ArrayList<AlignSeq>();

      for (SequenceI sq : ochains)
      {
        SequenceI bestm = null;
        AlignSeq bestaseq = null;
        int bestscore = 0;
        for (SequenceI msq : al.getSequences())
        {
          AlignSeq aseq = AlignSeq.doGlobalNWAlignment(msq, sq,
                  dnaOrProtein);
          if (bestm == null || aseq.getMaxScore() > bestscore)
          {
            bestscore = aseq.getMaxScore();
            bestaseq = aseq;
            bestm = msq;
          }
        }
        System.out.println("Best Score for " + (matches.size() + 1) + " :"
                + bestscore);
        matches.add(bestm);
        aligns.add(bestaseq);
        al.deleteSequence(bestm);
      }
      for (int p = 0, pSize = seqs.size(); p < pSize; p++)
      {
        SequenceI sq, sp = seqs.get(p);
        int q;
        if ((q = ochains.indexOf(sp)) > -1)
        {
          seqs.set(p, sq = matches.get(q));
          sq.setName(sp.getName());
          sq.setDescription(sp.getDescription());
          sq.transferAnnotation(sp, aligns.get(q).getMappingFromS1(false));
          int inspos = -1;
          for (int ap = 0; ap < annotations.size();)
          {
            if (((AlignmentAnnotation) annotations.get(ap)).sequenceRef == sp)
            {
              if (inspos == -1)
              {
                inspos = ap;
              }
              annotations.remove(ap);
            }
            else
            {
              ap++;
            }
          }
          if (sq.getAnnotation() != null)
          {
            annotations.addAll(inspos, Arrays.asList(sq.getAnnotation()));
          }
        }
      }
    }
  }

  /**
   * make a friendly ID string.
   * 
   * @param dataName
   * @return truncated dataName to after last '/'
   */
  private String safeName(String dataName)
  {
    int p = 0;
    while ((p = dataName.indexOf("/")) > -1 && p < dataName.length())
    {
      dataName = dataName.substring(p + 1);
    }
    return dataName;
  }

  public void makeResidueList()
  {
    for (int i = 0; i < chains.size(); i++)
    {
      ((PDBChain) chains.elementAt(i)).makeResidueList();
    }
  }

  public void makeCaBondList()
  {
    for (int i = 0; i < chains.size(); i++)
    {
      ((PDBChain) chains.elementAt(i)).makeCaBondList();
    }
  }

  public PDBChain findChain(String id)
  {
    for (int i = 0; i < chains.size(); i++)
    {
      if (((PDBChain) chains.elementAt(i)).id.equals(id))
      {
        return (PDBChain) chains.elementAt(i);
      }
    }

    return null;
  }

  public void setChargeColours()
  {
    for (int i = 0; i < chains.size(); i++)
    {
      ((PDBChain) chains.elementAt(i)).setChargeColours();
    }
  }

  public void setColours(jalview.schemes.ColourSchemeI cs)
  {
    for (int i = 0; i < chains.size(); i++)
    {
      ((PDBChain) chains.elementAt(i)).setChainColours(cs);
    }
  }

  public void setChainColours()
  {
    for (int i = 0; i < chains.size(); i++)
    {
      ((PDBChain) chains.elementAt(i)).setChainColours(Color.getHSBColor(
              1.0f / (float) i, .4f, 1.0f));
    }
  }

  public boolean isRNA(SequenceI seqs)
  {
    for (int i = 0; i < seqs.getLength(); i++)
    {
      if ((seqs.getCharAt(i) != 'A') && (seqs.getCharAt(i) != 'C')
              && (seqs.getCharAt(i) != 'G') && (seqs.getCharAt(i) != 'U'))
      {
        return false;
      }
    }

    return true;

  }
}
